import rclpy
from rclpy.node import Node
from sensor_msgs.msg import LaserScan
from ackermann_msgs.msg import AckermannDriveStamped
import numpy as np
from dataclasses import dataclass
from typing import Optional

@dataclass
class PID:
    kp: float = 0.0
    ki: float = 0.0
    kd: float = 0.0
    setpoint: float = 0.0
    out_limits: tuple[Optional[float], Optional[float]] = (-1.0, 1.0)
    integral: float = 0.0
    previous_error: Optional[float] = None

    def reset(self) -> None:
        self.integral = 0.0
        self.previous_error = None

    def update(self, measurement: float, dt: Optional[float]) -> float:
        error: float = self.setpoint - measurement
        
        if dt and dt > 0.0:
            self.integral += error * dt
        
        derivative: float = (
            (error - self.previous_error) / dt
            if dt and dt > 0.0 and self.previous_error is not None
            else 0.0
        )
        
        u: float = self.kp * error + self.ki * self.integral + self.kd * derivative
        
        low, high = self.out_limits
        if low is not None and u < low:
            u = low
        if high is not None and u > high:
            u = high
        
        self.previous_error = error
        return u


def get_range(
    lidar_range_array: np.ndarray,
    angle: float,
    min_angle: float = -3.0 * np.pi / 4.0,
    max_angle: float = 3.0 * np.pi / 4.0,
):
    angle_increment = (max_angle - min_angle) / (lidar_range_array.size - 1)
    index = round((angle - min_angle) / angle_increment)
    
    if lidar_range_array is not None and index < lidar_range_array.size:
        return lidar_range_array[index]
    else:
        return 0.0


def get_distance_from_wall(lidar_range_array: np.ndarray):
    theta = np.radians(45.0)
    b = get_range(lidar_range_array, np.radians(90))
    a = get_range(lidar_range_array, np.radians(45))
    alpha = np.arctan2(a * np.cos(theta) - b, a * np.sin(theta))
    D = b * np.cos(alpha)
    error = D + 1.5 * np.sin(alpha)
    return error


class WallFollow(Node):
    def __init__(self):
        super().__init__("wall_follow")
        
        self.steering_pid = PID(kp=1.0, kd=0.0, ki=0, setpoint=-0.6)
        
        self.scan_sub = self.create_subscription(
            LaserScan, "/scan", self.scan_callback, 10
        )
        
        self.drive_pub = self.create_publisher(
            AckermannDriveStamped, "/drive", 10
        )

    def scan_callback(self, msg):
        print("test")
        
        speed = 1.0
        steering = self.steering_pid.update(
            -get_distance_from_wall(np.array(msg.ranges)),
            msg.time_increment,
        )
        
        ackermann_msg = AckermannDriveStamped()
        ackermann_msg.header.stamp = self.get_clock().now().to_msg()
        ackermann_msg.header.frame_id = "base_link"
        ackermann_msg.drive.steering_angle = float(steering)
        ackermann_msg.drive.speed = float(speed)
        
        self.drive_pub.publish(ackermann_msg)


def main(args=None):
    rclpy.init(args=args)
    wall_follow = WallFollow()
    
    try:
        rclpy.spin(wall_follow)
    except KeyboardInterrupt:
        pass
    finally:
        wall_follow.destroy_node()
        rclpy.shutdown()


if __name__ == "__main__":
    main()


from setuptools import find_packages, setup

package_name = "disparity_extender"

setup(
    name=package_name,
    version="0.0.0",
    packages=find_packages(exclude=["test"]),
    data_files=[
        ("share/ament_index/resource_index/packages", ["resource/" + package_name]),
        ("share/" + package_name, ["package.xml"]),
        ("share/" + package_name + "/launch", ["launch/disparity_extender.launch.py"]),
    ],
    package_data={"": ["py.typed"]},
    install_requires=["setuptools"],
    zip_safe=True,
    maintainer="root",
    maintainer_email="root@todo.todo",
    description="TODO: Package description",
    license="TODO: License declaration",
    extras_require={
        "test": [
            "pytest",
        ],
    },
    entry_points={
        "console_scripts": [
            "disparity_extender = disparity_extender.disparity_extender:main"
        ],
    },
)



from setuptools import find_packages, setup

package_name = 'wall_follow'

setup(
    name=package_name,
    version='0.0.0',
    packages=find_packages(exclude=['test']),
    data_files=[
        ('share/ament_index/resource_index/packages',
            ['resource/' + package_name]),
        ('share/' + package_name, ['package.xml']),
        ('share/' + package_name + '/launch', ['launch/wall_follow.launch.py']),
    ],
    package_data={'': ['py.typed']},
    install_requires=['setuptools'],
    zip_safe=True,
    maintainer='Alistair Keiller',
    maintainer_email='alistair@keiller.net',
    description='TODO: Package description',
    license='TODO: License declaration',
    extras_require={
        'test': [
            'pytest',
        ],
    },
    entry_points={
        'console_scripts': [
            'wall_follow = wall_follow.wall_follow:main',
        ],
    },
)



import gymnasium as gym
import rustoracerpy

# Initialise the environment
env = gym.make("Rustoracer-v0", yaml="maps/berlin.yaml")

# Reset the environment to generate the first observation
observation, info = env.reset(seed=42)
for _ in range(1000):
    # this is where you would insert your policy
    action = env.action_space.sample()

    # step (transition) through the environment with the action
    # receiving the next observation, reward and if the episode has terminated or truncated
    observation, reward, terminated, truncated, info = env.step(action)

    # If the episode has ended then we can reset to start a new episode
    if terminated or truncated:
        observation, info = env.reset()

env.close()



import rclpy
from rclpy.node import Node
from sensor_msgs.msg import LaserScan
from std_msgs.msg import Float32
from ackermann_msgs.msg import AckermannDriveStamped
import numpy as np
from typing import Any

# Constants
min_angle: float = -np.pi / 2.0  # radians
max_angle: float = np.pi / 2.0  # radians
bubble_size: int = 120  # lidar points


def index_to_angle(index: int, num_points: int) -> float:
    angle_increment = (max_angle - min_angle) / (num_points - 1)
    angle = min_angle + index * angle_increment
    return angle


def find_best_point(lidar_range_array: np.ndarray) -> int:
    best_index = 0
    best_min_distance = 0.0

    for i in range(len(lidar_range_array) - bubble_size + 1):
        window = lidar_range_array[i : i + bubble_size]
        min_distance = np.min(window)

        if min_distance > best_min_distance:
            best_min_distance = min_distance
            best_index = i

    return best_index + bubble_size // 2


def compute_speed(target_distance: float) -> float:
    if target_distance < 6.0:
        return target_distance / 6.0
    return 1.0


class DisparityExtender(Node):
    def __init__(self):
        super().__init__("disparity_extender")
        self.scan_sub = self.create_subscription(
            LaserScan, "/scan", self.scan_callback, 10
        )

        self.drive_pub = self.create_publisher(
            AckermannDriveStamped, "/drive", 10
        )

    def scan_callback(self, msg):
        lidar_range_array = np.array(msg.ranges)

        sixth = lidar_range_array.size // 6
        lidar_range_array = lidar_range_array[sixth:-sixth]

        best_point_index = find_best_point(lidar_range_array)

        best_point_angle = index_to_angle(best_point_index, lidar_range_array.size)
        steering = best_point_angle / (np.pi / 2.0)

        target_distance = lidar_range_array[best_point_index]
        speed = compute_speed(target_distance)

        ackermann_msg = AckermannDriveStamped()
        ackermann_msg.header.stamp = self.get_clock().now().to_msg()
        ackermann_msg.header.frame_id = "base_link"
        ackermann_msg.drive.steering_angle = float(steering)
        ackermann_msg.drive.speed = float(speed)
        
        self.drive_pub.publish(ackermann_msg)


def main(args=None):
    rclpy.init(args=args)

    disparity_extender = DisparityExtender()

    try:
        rclpy.spin(disparity_extender)
    except KeyboardInterrupt:
        pass
    finally:
        disparity_extender.destroy_node()
        rclpy.shutdown()


if __name__ == "__main__":
    main()



from launch import LaunchDescription
from launch_ros.actions import Node


def generate_launch_description():
    return LaunchDescription(
        [
            Node(
                package="disparity_extender",
                executable="disparity_extender",
            ),
        ]
    )



from launch import LaunchDescription
from launch_ros.actions import Node

def generate_launch_description():
   return LaunchDescription([
    Node(
        package='wall_follow',
        executable='wall_follow',
    ),
   ])


from __future__ import annotations

import gymnasium as gym
import numpy as np
from gymnasium import spaces

from rustoracerpy.rustoracer import PySim


class RustoracerEnv(gym.Env):
    def __init__(self, yaml: str, init_pose=(0.0, 0.0, 0.0), max_steps=10_000):
        super().__init__()
        self._sim = PySim(yaml)
        self._pose = list(init_pose)
        self._max_steps = max_steps
        self._steps = 0

        self.observation_space = spaces.Box(
            0.0, self._sim.max_range, shape=(self._sim.n_beams,), dtype=np.float64
        )
        self.action_space = spaces.Box(
            np.array([-0.4189, 0.4189]), np.array([0.0, 10.0]), dtype=np.float64
        )

    def reset(self, *, seed=None, options=None):
        super().reset(seed=seed)
        self._steps = 0
        scan, pose, _ = self._sim.reset(self._pose)
        return scan, {"pose": pose}

    def step(self, action):
        self._steps += 1
        scan, pose, col = self._sim.step(float(action[0]), float(action[1]))
        reward = -100.0 if col else float(action[1])
        return scan, reward, col, self._steps >= self._max_steps, {"pose": pose}

    def close(self):
        pass
